@@ -19,6 +19,8 @@ GType vte_terminal_cursor_blink_mode_get_type (void);
 #define VTE_TYPE_TERMINAL_CURSOR_BLINK_MODE (vte_terminal_cursor_blink_mode_get_type ())
 GType vte_terminal_cursor_shape_get_type (void);
 #define VTE_TYPE_TERMINAL_CURSOR_SHAPE (vte_terminal_cursor_shape_get_type ())
+GType vte_terminal_write_flags_get_type (void);
+#define VTE_TYPE_TERMINAL_WRITE_FLAGS (vte_terminal_write_flags_get_type ())
 GType vte_terminal_anti_alias_get_type (void);
 #define VTE_TYPE_TERMINAL_ANTI_ALIAS (vte_terminal_anti_alias_get_type ())
 G_END_DECLS
